#!/usr/bin/python
"""WSGI server example"""
from __future__ import print_function
from gevent.pywsgi import WSGIServer

def application(environ, start_response):
	start_response('200 OK', [('Content-Type', 'text/plain')])
	#f = open ("/var/local/skitaid-pub/test/static/test.htm", "rb")	
	#data = f.read ()
	#f.close ()
	data = b"pong"
	return [data]

print('Serving on 5001...')	
WSGIServer(('0.0.0.0', 5001), application, spawn =10000).serve_forever()



try:
	import xmlrpc.client as xmlrpclib
except ImportError:
	import xmlrpclib
	
import base64
try: 
	from urllib.parse import urlparse, quote
except ImportError:
	from urllib import quote
	from urlparse import urlparse
				
from skitai.lib import producers, strutil


class XMLRPCRequest:
	content_type = "text/xml"
			
	def __init__ (self, uri, method, params = (), headers = None, encoding = "utf8", auth = None, logger = None):
		self.uri = uri
		self.method = method
		self.params = params
		self.headers = headers
		self.encoding = encoding
		self.auth = (auth and type (auth) is not tuple and tuple (auth.split (":", 1)) or auth)
		self.logger = logger
		self.address, self.path = self.split (uri)
		self.__xmlrpc_serialized = False
		self.data = self.serialize ()		
	
	def xmlrpc_serialized (self):
		return self.__xmlrpc_serialized
		
	def set_address (self, address):
		self.address = address
		
	def get_address (self):
		return self.address
		
	def get_method (self):
		return "POST"
		
	def split (self, uri):
		if uri.find ("://") == -1:
			return None, uri
			
		scheme, address, script, params, qs, fragment = urlparse (uri)		
		if not script: script = "/"
		path = script
		if params: path += ";" + params
		if qs: path += "?" + qs
		
		try: 
			host, port = address.split (":", 1)
			port = int (port)
		except ValueError:
			host = address
			if scheme in ("http", "ws"):
				port = 80
			else:
				port = 443	
		
		return (host, port)	, path
		
	def serialize (self):
		self.__xmlrpc_serialized = True
		return xmlrpclib.dumps (self.params, self.method, encoding=self.encoding, allow_none=1).encode ("utf8")
	
	def get_auth (self):
		return self.auth
		
	def get_data (self):
		return self.data
	
	def get_useragent (self):
		return "Mozilla/5.0 (compatible; Skitaibot/0.1a)"
				
	def get_content_type (self):
		if self.headers:
			for k, v in list(self.headers.items ()):
				if k.lower () == "content-length":
					del self.headers [k]
				elif k.lower () == "content-type":
					self.content_type = v
					del self.headers [k]
		return self.content_type
	
	def get_headers (self):
		if self.headers:
			return list(self.headers.items ())
		else:
			return []	
			
	
class HTTPRequest (XMLRPCRequest):
	content_type = "application/x-www-form-urlencoded; charset=utf-8"
	def get_method (self):
		return self.method.upper ()
					
	def serialize (self):
		# formdata type can be string, dict, boolean
		if not self.params:
			# no content, no content-type
			self.content_type = None		
			return b""

		if type (self.params) is type ({}):
			if self.get_content_type () != "application/x-www-form-urlencoded":
				raise TypeError ("POST Body should be string or can be encodable")
			fm = []
			for k, v in list(self.params.items ()):
				if self.encoding:
					v = v.decode (self.encoding)					
				fm.append ("%s=%s" % (quote (k), quote (v)))
			return "&".join (fm).encode ("utf8")
		
		if strutil.is_encodable (self.params):
			return self.params.encoding ("utf8")
		
		if self.encoding and strutil.is_decodable (self.params):
			return self.params.decode (self.encoding).encoding ("utf8") 
			
		return self.params
		
		
class HTTPPutRequest (HTTPRequest):
	# PUT content-type hasn't got default type
	content_type = None
			
	def get_method (self):
		return "PUT"
					
	def serialize (self):
		if type (self.params) is not str:
			raise TypeError ("PUT body must be string")
		if self.encoding:
			return self.params.decode (self.encoding).encode ("utf8")
		return self.params.encode ("utf8")
		
		
class HTTPMultipartRequest (HTTPRequest):
	boundary = "-------------------SAE-20150614204358"
	
	def __init__ (self, uri, method, params = {}, headers = None, encoding = None, auth = None, logger = None):
		HTTPRequest.__init__ (self, uri, method, params, headers, encoding, auth, logger)
		if type (self.params) is bytes:
			self.find_boundary ()
	
	def get_method (self):
		return "POST"
						
	def get_content_type (self):
		HTTPRequest.get_content_type (self) # for remove content-type header
		return "multipart/form-data; boundary=" + self.boundary
			
	def find_boundary (self):
		s = self.params.find (b"\r\n")
		if s == -1:
			raise ValueError("boundary not found")
		b = self.params [:s]			
		if b [:2] != b"--":
			raise ValueError("invalid multipart/form-data")
		self.boundary = b [2:s]
		
	def serialize (self):
		if type (self.params) is type ({}):
			return producers.multipart_producer (self.params, self.boundary, self.encoding)
		return self.params




import base64
from hashlib import md5
import re
try:
	from urllib.parse import urlparse, quote_plus	, unquote_plus
except ImportError:
	from urlparse import urlparse
	from urllib import quote_plus, unquote_plus	
import time
from skitai.protocol.http import util

DEBUG = False
if not DEBUG:
	from . import localstorage

MAX_MOVED = 5
KEYWORDS = ('from', "label", "with")
LONGOPTS = ('cookie', 'user-agent', 'version', 'auth', 'proxy', 'tunnel', 'content-type', 'connection', 'opcode')
HEARES_NOT_ALLOWED = ("cookie", "content-type", "connection", "referer")
METHODS = ("get", "post", "head", "put", "delete", "options", "trace", "connect", "upload")
DEFAULT_PORT_MAP = {'http': 80, 'https': 443, 'ws': 80, 'wss': 443}
NON_IP_COMPONENT = re.compile ("[^.0-9]")

class UserData:
	def __getattr__ (self, n):
		return None
	
class HTTPOptions (UserData):
	version = "1.1"
	connection = "close"
	user_agent = "Mozilla/5.0 (compatible; Skitaibot/0.1a)"
	headers = {"Cache-Control": "max-age=0"}

class Attributes (UserData):
	depth = 0
	retrys = 0
	moved = []

	
class RQL:	
	def __init__ (self, rql):
		self.rql = rql
		self.udata = UserData ()
		self.hconf = HTTPOptions ()
		self.uinfo = Attributes ()
		self.__verson_changed = False
		self.parse ()
	
	def __str__ (self):
		return self.uinfo.rfc
	
	def get_headers (self):
		h = list (self.hconf.headers.items ())
		if self.hconf.cookie:
			h.append (("Cookie", self.hconf.cookie))
		if self.uinfo.referer:
			h.append (("Referer", self.uinfo.referer))
		return dict (h)
	
	def check_end_quote (self, quote, token):
		if token and token [-1] == quote and (len (token) == 1 or token [-2] != "\\"):
			token = token [:-1].replace ("\\" + quote, quote)
			quote = None
		else:	
			token = token.replace ("\\" + quote, quote)
		return quote, token
		
	def parse (self):
		if not self.rql.strip ():
			raise AssertionError ("Empty RQL")
		
		d = {}
		current_key = None
		current_quote = None
		
		for token in self.rql.split(" "):
			if not token and (current_key is None or not d [current_key]):
				continue
			
			if current_quote is not None:
				current_quote, token = self.check_end_quote (current_quote, token)
				d[current_key] += ' ' + token
				continue
				
			tk = token.lower()
			if tk in KEYWORDS or tk in METHODS:				
				current_key = tk
				d [current_key] = ''
			
			elif tk [:2] in ("--"):
				got_long_opt = True
				current_key = token				
				d[current_key] = ''
					
			else:
				try: 
					if d[current_key]:
						d[current_key] += ' ' + token
												
					else:
						if token [0] in "\"'":
							current_quote = token [0]
							token = token [1:]
							current_quote, token = self.check_end_quote (current_quote, token)
						d[current_key] = token
							
				except KeyError:
					current_key = 'get'					
					d[current_key] = token
		
		if current_quote:
			raise AssertionError ("Quotes Not Matched")
			
		self.analyze (d)
		self.set_attributes ()
		
	def set_referer (self, url, page_id = None):
		self.uinfo.referer = url
		self.uinfo.referer_id = page_id		
			
	def analyze (self, d):
		for k, v in d.items ():
			v = v.strip ()
			if k [:2] == "--":
				if k.startswith ("--head-"):
					hkey = k [7:]
					if hkey.lower () in HEARES_NOT_ALLOWED:
						raise AssertionError ("Not Allowed '%s' Header" % hkey)
					self.hconf.headers [hkey] = v
				
				elif k [2:] in LONGOPTS:
					# LONGOPTS = ('cookie', 'user-agent', 'version', 'auth', 'proxy', 'tunnel', 'content-type', 'connection')
					okey = k [2:]
					if okey == "cookie":
						if not DEBUG:
							localstorage.g.set_cookie_from_data (self.uinfo.rfc, v)
					elif okey == "user-agent":
						self.hconf.user_agent = v					
					elif okey == "content-type":
						self.hconf.content_type = v
					elif okey == "connection":
						self.hconf.connection = v
					else:
						setattr (self.hconf, okey, v)
				
				else:
					if k [-4:] == ":int":
						setattr (self.udata, k [2:-4], int (v))
					else:
						setattr (self.udata, k [2:], v)
							
			else:	
				if k in METHODS:
					self.uinfo.method = k
					if v.find ("://") == -1:
						v = "http://" + v # assume
					self.uinfo.url = v					
				elif k == "label":
					self.uinfo.label = v
				elif k == "from":
					self.set_referer (v)					
				elif k == "with":	
					self.uinfo.data = v
		
	def get_header (self, name, default = None):
		name = name.lower ()
		for k, v in self.hconf.headers.items ():
			if k.lower () == name:
				return v
		return default
		
	def set_attributes (self):
		url = self.uinfo.url
		self.uinfo.scheme, self.uinfo.netloc, self.uinfo.script, self.uinfo.params, self.uinfo.querystring, self.uinfo.fragment = urlparse (url)
		if self.uinfo.data and self.uinfo.method not in ("post", "put"):
			raise ValueError ("Form exists but method isn't post or get")
		if self.uinfo.method in ("post", "put") and not self.uinfo.data:
			raise ValueError ("No form data")
		if not self.uinfo.data and self.hconf.content_type:
			raise ValueError ("Needn't content-type")
		if self.uinfo.data and self.uinfo.method == "post" and self.hconf.content_type is None:
			self.hconf.content_type = "application/x-www-form-urlencoded; charset=utf-8"		
			
		if not self.uinfo.script: 
			self.uinfo.script = '/'
		if self.uinfo.querystring:
			try: self.uinfo.querystring = uitl.strencode (self.uinfo.querystring)
			except: pass
		
		uri = self.uinfo.script #real request uri
		if self.uinfo.params:
			uri += ';' + self.uinfo.params
		if self.uinfo.querystring:
			uri += '?' + self.uinfo.querystring
		self.uinfo.uri = uri
		
		position = self.uinfo.netloc.find('@')
		if position > -1:
			self.hconf.auth, self.uinfo.netloc = self.uinfo.netloc.split ("@", 1)
		
		if self.hconf.auth:
			if self.get_header ("authorization"):
				raise ValueError ("Authorization Information Conflict")
				
			try: 
				self.hconf.username, self.hconf.password = self.hconf.auth.split (':', 1)
			except ValueError:	
				pass
			else:	
				self.hconf.auth = (self.hconf.username, self.hconf.password)

		try:
			self.uinfo.netloc, self.uinfo.port = self.uinfo.netloc.split (':', 1)
		except ValueError:
			try:
				self.uinfo.netloc, self.uinfo.port = self.uinfo.netloc, DEFAULT_PORT_MAP [self.uinfo.scheme]
			except KeyError:
				raise ValueError ("Unknown URL Scheme")
		
		self.uinfo.netloc = self.uinfo.netloc.lower ()
		try: self.uinfo.port = int (self.uinfo.port)
		except: self.uinfo.port = 80
		
		if NON_IP_COMPONENT.search (self.uinfo.netloc):
			netloc = self.uinfo.netloc
			netloc2 = netloc.split (".")
			self.uinfo.topdomain = netloc2 [-1]
			
			if len (netloc2) == 1:
				self.uinfo.host = ""	
				self.uinfo.domain = netloc
				self.uinfo.subdomain = netloc				
			elif len (netloc2) == 2:
				self.uinfo.host = ""	
				self.uinfo.domain = netloc
				self.uinfo.subdomain = netloc2 [0]				
			elif len (netloc2 [-1]) >= 3:
				self.uinfo.domain = ".".join (netloc2 [-2:])
				self.uinfo.host = ".".join (netloc2 [:-2])
				self.uinfo.subdomain = netloc2 [-2]			
				if not self.uinfo.host:
					self.uinfo.host = "www"
			elif len (netloc2 [-1]) == 2:
				if len (netloc2) == 2:
					self.uinfo.domain = ".".join (netloc2)
					self.uinfo.subdomain = netloc2 [0]
					self.uinfo.host = "www"
				elif len (netloc2) == 3:
					if netloc2 [0] == "www":
						self.uinfo.domain = ".".join (netloc2 [-2:])
						self.uinfo.subdomain = netloc2 [-2]
					else:
						self.uinfo.domain = ".".join (netloc2)
						self.uinfo.subdomain = netloc2 [0]
					self.uinfo.host = "www"
				else:
					self.uinfo.domain = ".".join (netloc2 [-3:])
					self.uinfo.host = ".".join (netloc2 [:-3])
					self.uinfo.subdomain = netloc2 [-3]
		
		if DEFAULT_PORT_MAP [self.uinfo.scheme] == self.uinfo.port:
			self.uinfo.rfc = '%s://%s%s' % (self.uinfo.scheme, self.uinfo.netloc, self.uinfo.uri)
		else:
			self.uinfo.rfc = '%s://%s:%d%s' % (self.uinfo.scheme, self.uinfo.netloc, self.uinfo.port, self.uinfo.uri)	
		self.uinfo.page_id = self.geneate_page_id ()
		self.uinfo.path_id = self.geneate_page_id (False)
		if not DEBUG and localstorage.g:
			self.hconf.cookie = localstorage.g.get_cookie_as_string (self.uinfo.rfc)		
		
	def to_version_11 (self):
		self.hconf.version = "1.1"
		if self.hconf.connection and self.hconf.connection.lower () == "close":
			del self.hconf.connection
		self.__verson_changed = True
	
	def inc_retrys (self):
		self.uinfo.retrys += 1
	
	def dec_retrys (self):
		self.uinfo.retrys -= 1
				
	def inherit (self, rql, moved = False):
		new = RQL (rql)
				
		if moved:
			if len (self.uinfo.moved) >= MAX_MOVED:
				raise AssertionError ("Maximum Moved Exceeded")
			new.uinfo.moved = self.uinfo.moved + [self.uinfo.page_id]
			if new.uinfo.page_id in new.uinfo.moved:
				raise AssertionError ("Moved Recursively")
		else:
			new.uinfo.depth = self.uinfo.depth + 1
					
		# inherit user-data
		for k in dir (self.udata):
			if k not in dir (new.udata): # prevent overwriting
				setattr (new.udata, k, getattr (self.udata, k))
		
		# inherit options
		for k in dir (self.hconf):
			if k in ("cookie", "content-type"):
				continue # no need, managed by local storage
			elif k == "headers":
				for k2, v2 in self.hconf.headers.items ():
					if k2 not in new.hconf.headers:
						new.hconf.headers [k2] = v2						
			elif k not in dir (new.hconf): # prevent overwriting
				setattr (new.hconf, k, getattr (self.hconf, k))
						
		if self.__verson_changed:
			new.to_version_11 ()
				
		# inherit attributes		
		new.set_referer (self.uinfo.rfc, self.uinfo.page_id)							
		return new
	
	def get_connection (self):
		return self.hconf.connection
				
	def get_useragent (self):
		return self.hconf.user_agent
	
	def __sort_args (self, data):
		args = {}
		for arg in data.split ("&"):
			try: k, v = arg.split ("=")
			except: continue
			if v:
				args [k] = v
						
		args = list(args.items ())
		args.sort ()
		
		argslist = []
		for k, v in args:
			argslist.append ("%s=%s" % (k, v))
		
		argslist.sort ()		
		return "&".join (argslist)

	def geneate_page_id (self, include_data = True):		
		signature = [
			self.uinfo.method,
			self.uinfo.netloc.startswith ("www.") and self.uinfo.netloc [4:] or self.uinfo.netloc, 
			str (self.uinfo.port), 
			self.uinfo.script
		]
		if include_data:
			signature.append (self.uinfo.params)
			if self.uinfo.querystring:
				signature.append (self.__sort_args (self.uinfo.querystring))
			if self.uinfo.data:
				signature.append (self.__sort_args (self.uinfo.data))
		return md5 (":".join (signature).encode ("utf8")).hexdigest ()
		
	def show (self):
		for d in ("udata", "uinfo", "hconf"):
			o = getattr (self, d)
			print (d, "\n======")
			for k in dir (o):
				if k [:2] == "__": continue
				print("-%s: %s" % (k, getattr (o, k)))
			print ()	
					

def make_page_id (url):
	return RQL (url).uinfo.page_id
make_pid = 	make_page_id

def make_path_id (url):
	return RQL (url).uinfo.path_id
	
def norm_space (s):
		return re.sub ("\s+", " ", s)	
		
def encode (s):	
	return quote_plus (s)
	
def decode (s):
	return unquote_plus (s)
				

if __name__ == "__main__":
	f = RQL ("from referer get http://url.com --useragent 'fire\\' fox --connection close' --header-Content-Type 'text/html' --id ''")
	f.show ()
	



# Web Socket message dump and load related codes are based on websocket-server 0.4
# by Johan Hanssen Seferidis
# https://pypi.python.org/pypi/websocket-server
#
# 2016. 1. 16 Modified by Hans Roh

import sys
import struct
from . import wsgi_handler
from hashlib import sha1
from base64 import b64encode
import time
from skitai.server.http_response import catch
from skitai.server.threads import trigger
try:
	from urllib.parse import quote_plus
except ImportError:
	from urllib import quote_plus	
from skitai.server import utility
from skitai.lib import strutil
from skitai import version_info, was as the_was
try:
	from cStringIO import StringIO as BytesIO
except ImportError:
	from io import BytesIO
import threading
import zlib
import threading

FIN    = 0x80
OPCODE = 0x0f
MASKED = 0x80
PAYLOAD_LEN = 0x7f
PAYLOAD_LEN_EXT16 = 0x7e
PAYLOAD_LEN_EXT64 = 0x7f

OPCODE_CONTINUATION = 0x0
OPCODE_TEXT = 0x1
OPCODE_BINARY = 0x2
OPCODE_CLOSE = 0x8
OPCODE_PING = 0x9
OPCODE_PONG = 0xa


class WebSocket:
	collector = None
	producer = None
	
	def __init__ (self, handler, request):		
		self.handler = handler		
		self.wasc = handler.wasc
		self.request = request		
		self.channel = request.channel
		self.channel.set_terminator (2)
		self.rfile = BytesIO ()
		self.masks = b""
		self.has_masks = True
		self.buf = b""
		self.payload_length = 0
		self.opcode = None
		self._closed = False
		
	def close (self):
		if self._closed: return
		self._closed = True
		if self.channel:
			self.channel.current_request = None  # break circ. ref
		self.handler.finish_request (self.request)		
	
	def closed (self):
		return self._closed
				
	def _tobytes (self, b):
		if sys.version_info[0] < 3:
			return map(ord, b)
		else:
			return b
		
	def collect_incoming_data (self, data):
		#print (">>>>", data)
		if not data:
			# closed connection
			self.close ()
			return
		
		if self.masks or (not self.has_masks and self.payload_length):
			self.rfile.write (data)			
		else:
			self.buf = data
	
	def found_terminator (self):
		#print ("-----", self.buf, self.opcode, self.payload_length, self.masks)
		if self.masks or (not self.has_masks and self.payload_length):
			# end of message
			masked_data = bytearray(self.rfile.getvalue ())
			if self.masks:
				masking_key = bytearray(self.masks)
				data = bytearray ([masked_data[i] ^ masking_key [i%4] for i in range (len (masked_data))])
			else:
				data = masked_data	
			
			if self.opcode == OPCODE_TEXT:
				# text
				data = data.decode('utf-8')
		
			self.payload_length = 0
			self.opcode = None
			self.masks = b""
			self.has_masks = True
			self.rfile.seek (0)
			self.rfile.truncate ()
			self.channel.set_terminator (2)
			self.handle_message (data)
		
		elif self.payload_length:
			self.masks = self.buf
			self.channel.set_terminator (self.payload_length)
		
		elif self.opcode:			
			if len (self.buf) == 2:
				fmt = ">H"
			else:	
				fmt = ">Q"
			self.payload_length = struct.unpack(fmt, self._tobytes(self.buf))[0]
			if self.has_masks:
				self.channel.set_terminator (4) # mask
			else:
				self.channel.set_terminator (self.payload_length)
		
		elif self.opcode is None:
			b1, b2 = self._tobytes(self.buf)
			fin    = b1 & FIN
			self.opcode = b1 & OPCODE
			#print (fin, self.opcode)
			if self.opcode == OPCODE_CLOSE:
				self.close ()
				return
				
			mask = b2 & MASKED
			if not mask:
				self.has_masks = False
			
			payload_length = b2 & PAYLOAD_LEN
			if payload_length == 0:
				self.opcode = None
				self.has_masks = True
				self.channel.set_terminator (2)
				return
			
			if payload_length < 126:
				self.payload_length = payload_length
				if self.has_masks:
					self.channel.set_terminator (4) # mask
				else:
					self.channel.set_terminator (self.payload_length)
			elif payload_length == 126:
				self.channel.set_terminator (2)	# short length
			elif payload_length == 127:
				self.channel.set_terminator (8) # long length
		
		else:
			raise AssertionError ("Web socket frame decode error")
			
	def send (self, message, op_code = OPCODE_TEXT):		
		header  = bytearray()
		if strutil.is_encodable (message):
			payload = message.encode ("utf8")
		else:
			payload = message
		payload_length = len(payload)

		# Normal payload
		if payload_length <= 125:
			header.append(FIN | op_code)
			header.append(payload_length)

		# Extended payload
		elif payload_length >= 126 and payload_length <= 65535:
			header.append(FIN | op_code)
			header.append(PAYLOAD_LEN_EXT16)
			header.extend(struct.pack(">H", payload_length))

		# Huge extended payload
		elif payload_length < 18446744073709551616:
			header.append(FIN | op_code)
			header.append(PAYLOAD_LEN_EXT64)
			header.extend(struct.pack(">Q", payload_length))
						
		else:
			raise AssertionError ("Message is too big. Consider breaking it into chunks.")
		
		m = header + payload
		trigger.wakeup (lambda p=self.channel, d=m: (p.push (d),))
	
	def handle_message (self, msg):
		raise NotImplementedError ("handle_message () not implemented")
		
	
class WebSocket1 (WebSocket):
	# WEBSOCKET_REQDATA			
	def __init__ (self, handler, request, apph, env, param_name):
		WebSocket.__init__ (self, handler, request)
		self.apph = apph
		self.env = env
		self.querystring = self.env.get ("QUERY_STRING", "")
		if self.querystring:
			self.querystring += "&"
		self.querystring += "%s=" % param_name
		
	def handle_message (self, msg):
		self.env ["QUERY_STRING"] = self.querystring + quote_plus (msg)		
		if self.env ["wsgi.multithread"]:
			args = (self.request, self.apph, (self.env, self.send), self.wasc.logger)
			self.wasc.queue.put (Job1 (*args))
		else:
			Job1 (*args) ()

class Job1 (wsgi_handler.Job):
	def exec_app (self):
		was = the_was._get ()
		was.request = self.request
		self.args [0]["skitai.was"] = was
				
		try:
			content = self.apph (*self.args) [0]
		except:
			content = self.apph.debug and "[ERROR]" + catch (0) or "[ERROR]"
		self.args [1] (content)


class WebSocket2 (WebSocket):
	# WEBSOCKET_DEDICATE if lock is None, or WEBSOCKET_MULTICAST
	def __init__ (self, handler, request):
		WebSocket.__init__ (self, handler, request)
		self.cv = threading.Condition (threading.RLock ())
		self.messages = []		
				
	def close (self):
		WebSocket.close (self)
		self.cv.acquire()
		self.cv.notify ()
		self.cv.release ()
		
	def getswait (self, timeout = 10):
		if self._closed:
			return None # closed channel
		self.cv.acquire()
		while not self.messages and not self._closed:
			self.cv.wait(timeout)
		if self._closed:
			self.cv.release()
			return None
		messages = self.messages
		self.messages = []
		self.cv.release()
		return messages
	
	def handle_message (self, msg):		
		self.cv.acquire()
		self.messages.append (msg)
		self.cv.notify ()
		self.cv.release ()

class WebSocket4 (WebSocket2):
	# WEBSOCKET_DEDICATE_THREADSAFE
	def __init__ (self, handler, request):
		WebSocket2.__init__ (self, handler, request)
		self.lock = threading.Lock ()
		
	def send (self, message, op_code = OPCODE_TEXT):
		self.lock.acquire ()
		try:
			WebSocket2.send (self, message, op_code)
		finally:
			self.lock.release ()
			
class Job2 (Job1):
	def handle_error (self):
		pass
		
	def exec_app (self):
		was = the_was._get () # create new was, cause of non thread pool
		was.request = self.request
		self.args [0]["skitai.was"] = was
		try:
			self.apph (*self.args)
		except:
			self.handle_error ()
			self.logger.trace ("app")
		the_was._del () # remove


class WebSocket3 (WebSocket):
	# WEBSOCKET_MULTICAST
	def __init__ (self, handler, request, server):
		WebSocket.__init__ (self, handler, request)
		self.client_id = request.channel.channel_number
		self.server = server
	
	def handle_message (self, msg):
		self.server.handle_message (self.client_id, msg)
	
	def close (self):
		WebSocket.close (self)
		self.server.handle_close (self.client_id)


class Job3 (Job2):
	def __init__(self, server, request, apph, args, logger):
		self.server = server
		Job2.__init__(self, request, apph, args, logger)
		
	def handle_error (self):
		self.server.close ()
		

class WebSocketServers:
	def __init__ (self):
		self.lock = threading.RLock ()
		self.wss = {}
	
	def get (self, gid):
		self.lock.acquire ()
		wss = self.wss [gid]
		self.lock.release ()
		return wss
		
	def has_key (self, gid):	
		self.lock.acquire ()
		has = gid in self.wss
		self.lock.release ()
		return has
		
	def create (self, gid):	
		self.lock.acquire ()
		wss = WebSocketServer (gid)
		self.wss [gid] = wss
		self.lock.release ()
		return wss
	
	def remove (self, gid):
		self.lock.acquire ()
		try: 
			del self.wss [gid]			
		except KeyError: 
			pass	
		self.lock.release ()		
	
	def close (self):
		self.lock.acquire ()
		for s in list (self.wss.values ()):
			s.close ()
		self.lock.release ()

websocket_servers = WebSocketServers ()
			
class WebSocketServer (WebSocket2):
	def __init__ (self, gid):
		self._closed = False
		self.gid = gid
		self.lock = threading.RLock ()
		self.cv = threading.Condition ()
		self.messages = []
		self.clients = {}
		
	def add_client (self, ws):
		self.lock.acquire ()
		self.clients [ws.client_id] = ws
		self.lock.release ()
		self.handle_message (ws.client_id, 1)
	
	def handle_close (self, client_id):
		self.lock.acquire ()
		try: del self.clients [client_id]
		except KeyError: pass		
		self.lock.release ()
		self.handle_message (client_id, -1)
		
	def handle_message (self, client_id, msg):
		self.cv.acquire()
		self.messages.append ((client_id, msg))
		self.cv.notifyAll ()
		self.cv.release ()
			
	def close (self):
		websocket_servers.remove (self.gid)
		self.clients = {}
		self.cv.acquire()
		self._closed = True
		self.cv.notifyAll ()
		self.cv.release ()
		
	def send (self, *args, **karg):
		raise AssertionError ("Can't use send() on WEBSOCKET_MULTICAST spec, use send_to(client_id, msg, op_code)")
	
	def sendto (self, client_id, msg, op_code = OPCODE_TEXT):		
		self.lock.acquire ()
		try:
			client = self.clients [client_id]
		except KeyError:
			client = None	
		self.lock.release ()
		if client:
			client.send (msg, op_code)
		
	def sendall (self, msg, op_code = OPCODE_TEXT):
		self.lock.acquire ()
		clients = list (self.clients.keys ())
		self.lock.release ()
		for client_id in clients:
			self.sendto (client_id, msg)
	

class Handler (wsgi_handler.Handler):
	def match (self, request):
		connection = request.get_header ("connection")
		return connection and connection.lower ().find ("upgrade") != -1 and request.version == "1.1" and request.command == "get"
	
	def close (self):
		websocket_servers.close ()
	
	GUID = '258EAFA5-E914-47DA-95CA-C5AB0DC85B11'.encode ()
	def calculate_response_key (self, key):		
		hash = sha1(key.encode() + self.GUID)
		response_key = b64encode(hash.digest()).strip()
		return response_key.decode('ASCII')
		
	def handle_request (self, request):				
		def donot_response (self, *args, **kargs):
			def push (thing):
				raise AssertionError ("Websocket can't use start_response ()")
			return push	
		
		origin = request.get_header ("origin")
		host = request.get_header ("host")
		protocol = request.get_header ("http_sec_websocket_protocol", 'unknown')
		securekey = request.get_header ("sec-webSocket-key")
		
		if not origin or not host or not securekey: 
			return request.response.error (400)
		
		if not (host.startswith ("localhost:") or origin.find (host) != -1 or origin == "null"):
			return request.response.error (403)
		
		path, params, query, fragment = request.split_uri ()		
		has_route = self.apps.has_route (path)
		if type (has_route) is int:
			return request.response.error (404)
		
		apph = self.apps.get_app (path)
		if not self.authorized (apph.get_callable(), request, has_route):
			return
			
		env = self.build_environ (request, apph)
		was = the_was._get ()
		was.request = request
		env ["skitai.was"] = was
		env ["websocket_init"] = ""
		# app should reply  (design type one of (1,2,3), keep-alive seconds)
		# when env has 'skitai.websocket_init'
		try:
			apph (env, donot_response)
			try:
				design_spec, keep_alive, param_name = env ["websocket_init"]
				del env ["websocket_init"]
			except (IndexError, ValueError): 
				raise AssertionError ("You should return (design_spec, keep_alive, param_name) where env has key 'skitai.websocket_init'")				
			assert design_spec in (1,2,3,4), "design_spec  should be one of (WEBSOCKET_REQDATA, WEBSOCKET_DEDICATE, WEBSOCKET_DEDICATE_THREADSAFE, WEBSOCKET_MULTICAST)"			
		except:
			self.wasc.logger.trace ("server",  request.uri)
			return request.response.error (500, why = apph.debug and catch (1) or "")
		
		header = [
			("Sec-WebSocket-Accept", self.calculate_response_key (securekey)),
			("Upgrade", "Websocket"),
			("Connection", "Upgrade"),
      ("WebSocket-Protocol", protocol),
      ("WebSocket-Location", "ws://" + host + path)
		]
		request.response.start_response ("101 Web Socket Protocol Handshake", header)
		request.response.done ()
		
		if design_spec == 1: 
			# WEBSOCKET_REQDATA			
			# Like AJAX, simple request of client, simple response data
			# the simplest version of stateless HTTP protocol using basic skitai thread pool
			ws = WebSocket1 (self, request, apph, env, param_name)
			env ["websocket"] = ws		
			self.channel_config (request, ws, keep_alive)
		
		elif design_spec in (2, 4): 
			# WEBSOCKET_DEDICATE 			
			# 1:1 wesocket:thread
			# Be careful, it will be consume massive thread resources
			if design_spec == 2:
				ws = WebSocket2 (self, request)
			else:
				ws = WebSocket4 (self, request)
			request.channel.add_closing_partner (ws)
			env ["websocket"] = ws
			self.channel_config (request, ws, keep_alive)
			job = Job2 (request, apph, (env, donot_response), self.wasc.logger)
			threading.Thread (target = job).start ()
		
		else: 
			# WEBSOCKET_MULTICAST
			# /chat?roomid=456, 
			# return (WEBSOCKET_MULTICAST, 600, "roomid")
			# websocketserver thread will be created by roomid
			# can send to all clients of group / specific client
			if not param_name:
				gidkey = path
			else:	
				gid = utility.crack_query (query).get (param_name, None)
				try:
					assert gid, "%s value can't find" % param_name
				except:
					self.wasc.logger.trace ("server",  request.uri)
					return request.response.error (500, why = apph.debug and catch (1) or "")
				gid = "%s/%s" % (path, gid)
			
			if not websocket_servers.has_key (gid):
				server = websocket_servers.create (gid)
				request.channel.add_closing_partner (server)
				env ["websocket"] = server
				job = Job3 (server, request, apph, (env, donot_response), self.wasc.logger)
				threading.Thread (target = job).start ()	
			
			server = websocket_servers.get (gid)				
			ws = WebSocket3 (self, request, server)
			server.add_client (ws)
			self.channel_config (request, ws, keep_alive)			
		
	def finish_request (self, request):		
		if request.channel:
			request.channel.close_when_done ()
		
	def channel_config (self, request, ws, keep_alive):
		request.channel.current_request = ws
		request.channel.set_response_timeout (keep_alive)
		request.channel.set_keep_alive (keep_alive)
		


"""
2015. 12. 7 Hans Roh


ResponseContainer
	logger
	
	callback
	
	udata
	
	uinfo
		url
		scheme
		path
		script
		params
		querystring
		fragment
		method
		netloc
		domain
		host
		port
		auth
		rfc
		referer
		page_id
		referer_id
		depth
		moved
			
	request	
		version
		connection
		user_agent
		proxy
		header		
		content_type
		encoding
		body
		
	response
		version
		code
		msg
		connection
		header
		content_type
		charset
		body
	
	set_cookie (self, k, v)
	get_cookie (self, k)
	set_item (self, k, v)
	get_item (self, k)
	advance (self, surl)	
	sleep (self, timeout)
			
"""
has_lxml = True
try:
	from . import treebuilder
	import html5lib
except ImportError:
	has_lxml = False
import math		
import json
import time

try:
	import xmlrpclib
except ImportError:
	import xmlrpc.client as xmlrpclib
from . import localstorage
try:
	from urllib.parse import urljoin
except ImportError:
	from urlparse import urljoin	
	
class RCRequest:
	def __init__ (self, obj):
		self._header_cache = {}
		self.set (obj)
		
	def set (self, handler):	
		self.header = handler.header
		self.uri = handler.uri
		self.version = handler.http_version
		self.proxy = handler.request.rql.hconf.proxy
		self.connection = handler.request.rql.hconf.connection
		self.user_agent = handler.request.rql.hconf.user_agent
		
		self.body = handler.request.get_data ()
		self.content_type = handler.request.get_content_type ()
		self.encoding = handler.request.encoding
	
	def get_header_with_attr (self, header, default = None):
		d = {}
		v = self.get_header (header)
		if v is None:
			return default, d
			
		v2 = v.split (";")
		if len (v2) == 1:
			return v, d
		for each in v2 [1:]:
			try:
				a, b = each.strip ().split ("=", 1)
			except ValueError:
				a, b = each.strip (), None
			d [a] = b
		return v2 [0], d	
			
	def get_header (self, header, default = None):
		header = header.lower()
		hc = self._header_cache
		if header not in hc:
			h = header + ':'
			hl = len(h)
			for line in self.header:
				if line [:hl].lower() == h:
					r = line [hl:].strip ()
					hc [header] = r
					return r
			hc [header] = None
			return default
		else:
			return hc[header] is not None and hc[header] or default


class RCResponse (RCRequest):
	def set (self, handler):	
		r = handler.response
		self.__baseurl = handler.request.rql.uinfo.rfc
		self.header = r.header
		self.version, self.code, self.msg = r.version, r.code, r.msg		
		self.content_type = None
		self.charset = None 
		self.__parser, self.__html, self.__etree = None, None, None
		
		ct = self.get_header ("content-type")
		if ct:
			ctl = ct.split (";")
			self.content_type = ctl [0]
			for param in ctl [1:]:
				if param.strip ().startswith ("charset="):
					self.charset = param.split ("=", 1)[-1].strip ().lower ()
					
		self.connection = self.get_header ("connection")				
		try:
			self.body = r.get_content ()
		except:
			handler.request.logger.trace ()
			self.body = b""
			self.code, self.msg = 720, "Response Content Error"
	
	def __len__ (self):
		return len (self.body)
	
	def __str__ (self):
		return self.string ()
			
	def html (self):
		global has_lxml		
		assert has_lxml is True, "missing lxml or html5lib"
		if self.__html: return self.__parser, self.__html		
		self.__parser, self.__html = treebuilder.Parser, treebuilder.html (self.body, self.__baseurl, self.charset)
		return self.__parser, self.__html
	
	def etree (self):
		global has_lxml		
		assert has_lxml is True, "missing lxml or html5lib"		
		if self.__etree: return self.__parser, self.__etree
		self.__parser, self.__etree = treebuilder.Parser, treebuilder.etree (self.body, self.charset)
		return self.__parser, self.__etree
			
	def binary (self):
		return self.body
	
	def string (self):
		return treebuilder.to_str (self.body, self.charset)
		
	def json (self):
		return json.loads (self.text ())
	
	def xmlrpc (self):
		return xmlrpclib.loads (self.text ())	
	
	def to_binary (self, s):
		if type (s) is str:
			return s.encode ("utf8")
		return s	
					
	def save_to (self, path, header = None, footer = None):
		if type (self.body) is None:
			return
			
		if type (self.body) is bytes:
			f = open (path, "wb")
			if header:
				f.write (self.to_binary (header))	
			f.write (self.body)
			if footer:
				f.write (self.to_binary (footer))
			f.close ()
			
		else:						
			raise TypeError ("Content is not bytes")
			
						
class ResponseContainer:
	def __init__ (self, handler, callback):
		self.__rql = handler.request.rql
		self.__asyncon = handler.asyncon
		
		self.uinfo = self.__rql.uinfo
		self.udata = self.__rql.udata
		self.hconf = self.__rql.hconf
		self.request = RCRequest (handler)
		self.response = RCResponse (handler)
		self.logger = handler.request.logger
		self.callback = callback
		
		for header in handler.response.get_header ():
			if header.lower ().startswith ("set-cookie: "):
				localstorage.g.set_cookie_from_string (
					self.uinfo.rfc,
					header [12:]
				)
		
	def set_cookie (self, k, v):
		localstorage.g.set_cookie (self.uinfo.rfc, k, v)
	
	def get_cookie (self, k):	
		localstorage.g.get_cookie (self.uinfo.rfc, k)
	
	def set_item (self, k, v):
		localstorage.g.set_item (self.uinfo.rfc, k, v)
	
	def get_item (self, k):	
		localstorage.g.get_item (self.uinfo.rfc, k)
		
	def stall (self, timeout):
		a, b = math.modf (timeout)
		for i in range (int (b)):
			self.__asyncon.set_event_time ()
			time.sleep (1)
		time.sleep (a)
	
	def resolve (self, url):
		return urljoin (self.uinfo.rfc, url)
	
	def inherit (self, surl):
		return self.__rql.inherit (surl)
		
	def relocate (self, url):
		from skitai import requests
		requests.add (self.__rql.inherit (self.resolve (url), True), self.callback, front = True)
		
	def visit (self, surl, callback = None):
		from skitai import requests
		requests.add (self.inherit (surl), callback and callback or self.callback)
	
	def retry (self):
		from skitai import requests
		self.uinfo.eurl.inc_retrys ()
		requests.add (self.__rql, self.callback, front = True)
	
	
	


import asynchat
import asyncore
import re
import os
import socket
import time
import sys
import zlib
import ssl
from errno import ECONNRESET, ENOTCONN, ESHUTDOWN, ECONNABORTED, EWOULDBLOCK
import select
import threading
from . import adns
from warnings import warn

DEBUG = False

class SocketPanic (Exception): pass
class TimeOut (Exception): pass

class AsynConnect (asynchat.async_chat):
	ac_in_buffer_size = 4096
	ac_out_buffer_size = 4096
	zombie_timeout = 10
	keep_alive_timeout = 10
	network_delay_timeout = 10	
	ready = None
	affluent = None	
	
	request_count = 0
	active = 0
	proxy = False
			
	def __init__ (self, address, lock = None, logger = None):
		self.address = address
		self.lock = lock
		self.logger = logger
		self._cv = threading.Condition ()		
		self.set_event_time ()
		self.proxy = False
		self.handler = None
		self.__history = []
		self.initialize_connection ()
		asynchat.async_chat.__init__ (self)
	
	def get_history (self):
		return self.__history
		
	def initialize_connection (self):		
		self._raised_ENOTCONN = 0 # for win32
		self._handshaking = False
		self._handshaked = False		
		
		self.established = False		
		self.upgraded = False		
		self.ready = None
		self.affluent = None		
			
	def set_event_time (self):
		self.event_time = time.time ()
		
	def set_proxy (self, flag = True):
		self.proxy = flag
	
	def is_proxy (self):
		return self.proxy
			
	def log (self, msg, logtype):
		if self.handler is not None and hasattr (self.handler, "log"):
			self.handler.log (msg, logtype)
		elif self.logger:
			self.logger (msg, logtype)
		else:
			warn ("No logger")
			
	def trace (self):		
		if self.handler is not None and hasattr (self.handler, "trace"):
			self.handler.trace ()
		elif self.logger:
			self.logger.trace ()
		else:
			warn ("No logger for traceback")
				
	def duplicate (self):
		return self.__class__ (self.address, self.lock, self.logger)
		
	def clean_shutdown_control (self, phase, time_in_this_phase):
		if phase == 2:
			self.handle_close (705, "Server Entered Shutdown Process")
		
	def readable (self):
		if self.affluent is not None:
			return asynchat.async_chat.readable (self)	and self.affluent ()
		return asynchat.async_chat.readable (self)	
	
	def writable (self):
		if self.ready is not None:
			return asynchat.async_chat.writable (self) and self.ready ()
		return asynchat.async_chat.writable (self)	
        
	def maintern (self, object_timeout):
		# check inconsistency, maybe impossible
		a, b = self.handler and 1 or 0, self.isactive () and 1 or 0
		if a != b:
			self.disconnect ()
			self.end_tran ()
					
		if time.time () - self.event_time > object_timeout:
			if not self.isactive ():	
				self.disconnect ()
				return True
		
		return False
	
	def is_channel_in_map (self, map = None):
		if map is None:
			map = self._map
		return self._fileno in map
		
	def set_active (self, flag, nolock = False):
		if DEBUG: self.__history.append ("SET ACTIVE %s" % flag) 
		if flag:
			flag = time.time ()
		else:
			flag = 0
			
		if nolock or self.lock is None:
			self.active = flag
			return
			
		self.lock.acquire ()
		self.active = flag
		self.request_count += 1
		self.lock.release ()
	
	def get_active (self, nolock = False):
		if nolock or self.lock is None:
			return self.active
		self.lock.acquire ()
		active = self.active
		self.lock.release ()	
		return active
	
	def isactive (self):	
		return self.get_active () > 0
	
	def isconnected (self):	
		return self.connected
		
	def get_request_count (self):	
		return self.request_count
	
	def add_channel (self, map = None):		
		if DEBUG: self.__history.append ("CHANNEL ADDED") 
		self.zombie_timeout =  self.network_delay_timeout		
		self._fileno = self.socket.fileno ()
		return asynchat.async_chat.add_channel (self, map)
		
	def del_channel (self, map = None):
		if DEBUG: self.__history.append ("CHANNEL REMOVED") 
		asynchat.async_chat.del_channel (self, map)
		# make default and sometimes reset server'stimeout	
		self.zombie_timeout =  self.keep_alive_timeout
				
	def create_socket (self, family, type):
		self.family_and_type = family, type
		sock = socket.socket (family, type)
		sock.setblocking (0)
		self.set_socket (sock)
	
	def connect (self):
		if adns.query:
			if DEBUG: self.__history.append ("QUERYING DNS")
			adns.query (self.address [0], "A", callback = self.continue_connect)
		else:
			# no adns query
			self.continue_connect (True)
		
	def continue_connect (self, answer = None):
		if not answer:
			if DEBUG: self.__history.append ("DNS FAILED")
			self.log ("DNS not found - %s" % self.address [0], "error")
			return self.handle_close (704, "DNS Not Found")
		
		if DEBUG: self.__history.append ("CREATING SOCKET & CONNECTING...") 
		self.initialize_connection ()
		self.create_socket (socket.AF_INET, socket.SOCK_STREAM)
		
		try:
			if not adns.query:				
				asynchat.async_chat.connect (self, self.address)
				
			else:	
				res = adns.get (self.address [0], "A")						
				if res:				
					ip = res [-1]["data"]
					if ip:
						asynchat.async_chat.connect (self, (ip, self.address [1]))																		
				else:
					self.handle_close (704, "DNS Not Found")
					
		except:	
			self.handle_error ()
	
	def recv (self, buffer_size):		
		self.set_event_time ()
		try:
			data = self.socket.recv(buffer_size)	
			#print ("+++++DATA", len (data), repr (data [:40]))				
			if not data:
				self.handle_close (700, "Connection closed unexpectedly in recv")
				return b''
			else:
				return data		
		except socket.error as why:
			if why.errno in asyncore._DISCONNECTED:				
				self.handle_close (700, "Connection closed unexpectedly in recv")
				return b''				
			else:
				raise
	
	def send (self, data):		
		self.set_event_time ()
		try:
			return self.socket.send (data)
		except socket.error as why:
			if why.errno == EWOULDBLOCK:
				return 0				
			elif why.errno in asyncore._DISCONNECTED:
				#print (">>>>>>>>>> why.errno == asyncore.ENOTCONN", why.errno == asyncore.ENOTCONN)
				if os.name == "nt" and why.errno == asyncore.ENOTCONN:
					# winsock sometimes raise ENOTCONN and sometimes recovered.
					# Found this error at http://file.hungryboarder.com:8080/HBAdManager/pa.html?siteId=hboarder&zoneId=S-2
					if self._raised_ENOTCONN <= 3:
						self._raised_ENOTCONN += 1
						return 0
					else:
						self._raised_ENOTCONN = 0
				self.handle_close (700, "Connection closed unexpectedly in send")
				return 0
			else:
				raise
	
	def close_if_over_keep_live (self):
		if time.time () - self.event_time > self.zombie_timeout:
			if DEBUG: self.__history.append ("KEEP-ALIVE TIMEOUT") 
			self.disconnect ()
	
	def initiate_send (self):
		if self.is_channel_in_map ():
			return asynchat.async_chat.initiate_send (self)
	
	def set_zombie_timeout (self, timeout = 10):
		self.zombie_timeout = timeout
	
	def set_zombie_timeout_by_case (self):
		if self.affluent or self.ready:
			self.zombie_timeout = self.network_delay_timeout * 2
		else:	
			self.zombie_timeout = self.network_delay_timeout
	
	def set_network_delay_timeout (self, timeout = 10):
		self.network_delay_timeout = timeout
	
	def set_keep_alive_timeout (self, timeout = 10):
		self.keep_alive_timeout = timeout
		
	def handle_connect (self):
		if DEBUG: self.__history.append ("CONNECTED") 
		try: 
			self.handler.has_been_connected ()
		except AttributeError:
			pass
		
	def handle_read (self):
		self.set_zombie_timeout_by_case ()		
		asynchat.async_chat.handle_read (self)
		
	def handle_write (self):
		self.set_zombie_timeout_by_case ()		
		asynchat.async_chat.handle_write (self)
			
	def handle_expt (self):
		self.logger ("socket panic", "fail")
		self.handle_close (703, "Socket Panic")
	
	def handle_error (self):
		self.trace ()
		self.handle_close(701, "Exception")
	
	def handle_timeout (self):
		self.log ("socket timeout", "fail")
		self.handle_close (702, "Socket Timeout")
		
	def handle_expt_event(self):
		err = self.socket.getsockopt(socket.SOL_SOCKET, socket.SO_ERROR)
		if err != 0:
			self.log ("Socket Error %d Occurred" % err, "warn")
			self.handle_close (706, "Socket %d Error" % err)
		else:
			self.handle_expt ()
	
	# proxy POST need no init_send
	def push (self, thing, init_send = True):
		if type (thing) is bytes:
			asynchat.async_chat.push (self, thing)
		else:
			self.push_with_producer (thing, init_send)	
		
	def push_with_producer (self, producer, init_send = True):
		self.producer_fifo.append (producer)
		if init_send:
			self.initiate_send ()
	
	def handle_close (self, code = 700, msg = "Disconnected by Server"):		
		if DEBUG: self.__history.append ("HANDLE_CLOSE %d %s" % (code, msg)) 
		self.errcode = code
		self.errmsg = msg
		self.close ()
							
	def collect_incoming_data (self, data):
		if not self.handler:
			self.logger ("recv data but no hander, droping data %d" % len (data), "warn")
			self.disconnect ()
			return
		self.handler.collect_incoming_data (data)
	
	def found_terminator (self):
		if not self.handler:
			self.logger ("found terminator but no handler", "warn")
			self.disconnect ()
			return # already closed
		self.handler.found_terminator ()
	
	def disconnect (self):
		# no error
		if DEBUG: self.__history.append ("DISCONNECTING") 
		self.handle_close (0, "")
	
	def reconnect (self):
		if DEBUG: self.__history.append ("RECONNECTING")
		self.disconnect ()
		self.connect ()
		
	def close (self):
		if self.socket:
			# self.socket is still None, when DNS not found
			asynchat.async_chat.close (self)
		
			# re-init asychat
			self.ac_in_buffer = b''
			self.incoming = []
			self.producer_fifo.clear()
			if DEBUG: self.__history.append ("DISCONNECTED")
		else:
			if DEBUG: self.__history.append ("CAN'T CLOSE, NOT CONNECTED")
		
		if self.handler is None:
			if DEBUG: self.__history.append ("NO HANDLER, GOTO END TRAN")
			self.end_tran () # automatic end_tran when timeout occured by maintern
		elif self.errcode:
			if DEBUG: self.__history.append ("CALL CONNECTION CLOSED")
			self.handler.connection_closed (self.errcode, self.errmsg)
			
	def end_tran (self):
		self.del_channel ()
		self.handler = None
		self.set_active (False)		
		if DEBUG: 
			self.__history.append ("END TRAN")
			self.__history = self.__history [-30:]
			
	def begin_tran (self, handler):
		self.errcode = 0
		self.errmsg = ""
			
		self.handler = handler		
		self.set_event_time ()
		if DEBUG: self.__history.append ("BEGIN TRAN %s %s" % (handler.method, handler.request.uri))
		
		if self.connected:
			self.close_if_over_keep_live () # check keep-alive
						
		try:
			if self.connected:
				# should keep order
				self.initiate_send ()
				self.add_channel ()
			else:
				self.connect ()
		except:
			self.handle_error ()
	
	
class AsynSSLConnect (AsynConnect):	
	ac_in_buffer_size = 65535 # generally safe setting 65535 for SSL
	ac_out_buffer_size = 65535
	
	def handshake (self):
		if not self._handshaking:
			err = self.socket.getsockopt(socket.SOL_SOCKET, socket.SO_ERROR)
			if err != 0:
				raise socket.error(err, _strerror(err))	
			self.socket = ssl.wrap_socket (self.socket, do_handshake_on_connect = False)
			self._handshaking = True
			
		try:
			self.socket.do_handshake ()
		except ssl.SSLError as why:
			if why.args [0] in (ssl.SSL_ERROR_WANT_READ, ssl.SSL_ERROR_WANT_WRITE):				
				return False
			raise ssl.SSLError(why)
		self._handshaked = True		
		return True
							
	def handle_connect_event (self):	
		if not self._handshaked and not self.handshake ():
			return
		# handshaking done
		self.handle_connect()
		self.connected = True
		
	def recv (self, buffer_size):
		self.set_event_time ()
		try:
			data = self.socket.recv (buffer_size)			
			if not data:				
				self.handle_close (700, "Connection closed unexpectedly")
				return b''
			else:				
				return data

		except ssl.SSLError as why:
			if why.errno == ssl.SSL_ERROR_WANT_READ:
				return b'' # retry			
			# closed connection
			elif why.errno in (ssl.SSL_ERROR_ZERO_RETURN, ssl.SSL_ERROR_EOF):
				self.handle_close (700, "Connection closed by SSL_ERROR_ZERO_RETURN or SSL_ERROR_EOF")
				return b''
				
			else:
				raise

	def send (self, data):
		self.set_event_time ()
		try:
			return self.socket.send (data)			

		except ssl.SSLError as why:
			if why.errno == ssl.SSL_ERROR_WANT_WRITE:
				return 0
			elif why.errno == ssl.SSL_ERROR_ZERO_RETURN:				
				self.handle_close (700, "Connection blosed by SSL_ERROR_ZERO_RETURN")
				return 0
			else:
				raise


class AsynSSLProxyConnect (AsynSSLConnect, AsynConnect):
	def handle_connect_event (self):
		if self.established:
			AsynSSLConnect.handle_connect_event (self)
		else:	
			AsynConnect.handle_connect_event (self)
	
	def recv (self, buffer_size):		
		if self._handshaked or self._handshaking:
			return AsynSSLConnect.recv (self, buffer_size)				
		else:
			return AsynConnect.recv (self, buffer_size)
			
	def send (self, data):		
		if self._handshaked or self._handshaking:
			return AsynSSLConnect.send (self, data)
		else:
			return AsynConnect.send (self, data)



#!/usr/bin/python

from flask import Flask, request, url_for
import os
from skitai import was

app = Flask(__name__)
app.debug = True
app.use_reloader = True

MULTIPART = """
<form action = "" enctype="multipart/form-data" method="post">
	<input type="hidden" name="submit-hidden" value="Genious">   
	<p></p>What is your name? <input type="text" name="submit-name" value="Hans Roh"></p>
	<p></p>What files are you sending? <br />
	<input type="file" name="file1"><br />
	<input type="file" name="file2">
	</p>
	<input type="submit" value="Send"> 
	<input type="reset">
</form>
"""

FORMDATA = """
<form action = "" method="post">
	<input type="hidden" name="submit-hidden" value="Genious">   
	<p></p>What is your name? <input type="text" name="submit-name" value="Hans Roh"></p>	
	<input type="submit" value="Send"> 
	<input type="reset">
</form>
"""
from flask import Response

@app.route('/stream')
def generate_large_csv():
	def generate():
		for row in range (100):
			yield str(row) + '\n'
	return Response(generate(), mimetype='text/plain')
    

@app.route('/was')
def wastest ():
	s = was.rpc ("http://210.116.122.187:3424/rpc2")
	s.bladese ("adsense.websearch", "computer", 0, 3)
	rs = s.getwait ()	
	return str (rs.data)

@app.route('/', methods=['GET', 'POST'])
def index ():	 
	if request.method == 'POST':		
		if request.files:
			file = request.files ['file1']
			filename = file.filename
			file.save (os.path.join("d:\\var\\upload", filename))
			return os.path.join("d:\\var\\upload", filename)
		else:
			return str (request.form)
	return FORMDATA+"<hr>"+MULTIPART+"<hr>URL_FOR: " + url_for ("generate_large_csv")

@app.route('/ping')
def ping ():
	return "pong"
	
if __name__ == '__main__':
    app.run(host="0.0.0.0", port=5002)




from __future__ import print_function
from skitai import requests
from skitai.lib import logger
import time


class timer:
	def __init__ (self):
		self.start = time.time()
	def end (self):
		return time.time() - self.start
				

def handle_response (rc):
	global total_sessions, clients, req, total_errors, resp_codes
	
	#print (rc.response.header)	
	total_sessions += 1
	try: resp_codes [rc.response.code] += 1
	except KeyError: resp_codes [rc.response.code] = 1
	if rc.response.code != 200: total_errors += 1
		
	print (total_sessions, end = " ")
	#print ("\r\n".join (rc.response.header))
	if total_sessions <= clients * req - clients:
		requests.add (url, handle_response)


def usage ():
	print ("%s [options] url" % sys.argv [0])
	print (" -c		concurrent clients (default: 1)")
	print (" -r		reuqests per clients (default: 1)")
	print (" -k		use keep-alive (default: no)")
	print (" --help		show help")
	

if __name__ == '__main__':
	import getopt, sys
	argopt, args = getopt.getopt (sys.argv[1:], "c:r:k", ["help"])
		
	clients = 1
	req = 1
	
	total_sessions = 0
	total_errors = 0
	resp_codes = {}
	use_keep_alive = False
	
	try:
		url = args [0]
	except IndexError:
		usage ()
		sys.exit ()	
	
	for k, v in argopt:
		if k == "-c": 
			clients = int (v)
		elif k == "-r":	
			req = int (v)
		elif k == "-k":	
			use_keep_alive = True
		elif k == "--help":		
			usage ()
			sys.exit ()
	
	requests.configure (
		logger.screen_logger (), 
		clients, 
		10, 
		default_option = "--http-connection " + (use_keep_alive and "keep-alive" or "close")
	)
	
	for i in range (clients):
		requests.add (url, handle_response)
	
	t = timer()
	requests.get_all ()
	total_time = t.end()
	
	print((
		'\n%d clients\n%d hits/client\n'
		'total hits:%d\n'
		'total errors:%d\n%.3f seconds\ntotal hits/sec:%.3f' % (
			clients,
			req,
			total_sessions,
			total_errors,
			total_time,
			total_sessions / total_time
		)
	))
						
	codes = list(resp_codes.items ())
	codes.sort (key = lambda x: x [1], reverse = True)
	print ("----------------------------\nresponse codes status")
	for each in codes:
		print ("%d: 	%d" % each)
